/**
 * Title:        Living Talks Web Conference Tool<p>
 * Description:  Tool using J2EE standards<p>
 * Copyright:    Copyright (c) Dr. Falk Langhammer<p>
 * Company:      Living Pages Research GmbH<p>
 * @author Falk Langhammer
 * @version 1.0
 */
package com.livis.livingtalks.ejb;

import com.livis.livingtalks.model.*;
import java.util.*;
import java.lang.reflect.*;
import java.rmi.*;
import javax.ejb.*;

public abstract class CachingEntityHandler extends EntityHandler implements DataHandler {

    // cache handling
    protected abstract WriteThroughCache sharedCache ();
    public void invalidateCache () { sharedCache ().invalidate (); }

    // override methods in EntityHandler
    public AbstractModel getShortmodel (String key) {
      return (AbstractModel) sharedCache ().read (key);
    }

    public void setModel (AbstractModel model) throws CreateException {
      sharedCache ().writeOld (model.getKey (), model);
    }

    public void remove (String key) throws RemoveException {
      try {
        sharedCache ().writeOld (key, null);
      } catch (LivingTalksCreateException le) {
        if (le.getNested () != null && le.getNested () instanceof RemoveException) throw (RemoveException) le.getNested ();
        else throw new RemoveException ("Item (" + key + ") could not be removed: " + le.getMessage () + "<br>\n" + le.getNested ());
      } catch (CreateException ce) {
        throw new RemoveException ("Item (" + key + ") could not be removed: " + ce.getMessage ());
      }
    }

    public String create (AbstractModel model) throws CreateException {
      return sharedCache ().writeNew (model.getKey (), model);
    }

    public Collection getShortmodelsAll () {
      if (!sharedCache ().isFilled ()) tryToFillCache ();
      if (!sharedCache ().isFilled ()) return super.getShortmodelsAll ();
      return sharedCache ().getAllData ();
    }

    public Collection getKeysAll () {
      if (!sharedCache ().isFilled ()) tryToFillCache ();
      if (!sharedCache ().isFilled ()) return super.getKeysAll ();
      return sharedCache ().getAllKeys ();
    }


    // implement DataHandler
    public String writeNew (String key, Object data) throws CreateException {
      return super.create ((AbstractModel) data);
    }

    public void writeOld (String key, Object data) throws CreateException {
      if (data == null) {
        try                        { super.remove (key); }
        catch (RemoveException re) { throw new LivingTalksCreateException (re); }
      }
      else super.setModel ((AbstractModel) data);
    }

    public Object read (String key) {
      return super.getShortmodel (key);
    }

    // implementation helpers
    /** attempt to fill the cache -- not guaranteed to work. Maybe some transaction locks data, or the cache lacks capacity */
    private void tryToFillCache () {
      try {
          sharedCache ().setFilled (false);
          Collection keys = super.getKeysAll ();
          for (Iterator i = keys.iterator (); i.hasNext (); )
            sharedCache ().read ((String) i.next ()); // touch 'em all
          sharedCache ().setFilled (true);
      } catch (Exception ex) {
        System.err.println ("Warning: Cache not filled, in class " + this.getClass ().getName () + ", reason is: \n" + ex);
      }
    }

    protected CachingEntityHandler (String homeName, Class homeClass) {
      super (homeName, homeClass);
    }
}

// these are very similiar to their non-caching counterparts, except for the last two lines
class CachingAuthorHandler extends CachingEntityHandler {
    public CachingAuthorHandler (String name) {
      super (name, AuthorHome.class);
      try {
        getterModel        = Author.    class.getMethod ("getModel",         null);
        setterModel        = Author.    class.getMethod ("putModel",         new Class [] { AuthorModel.class });
        creatorModel       = AuthorHome.class.getMethod ("create",           new Class [] { AuthorModel.class });
        finderByPrimaryKey = AuthorHome.class.getMethod ("findByPrimaryKey", new Class [] { String.class });
        finderAll          = AuthorHome.class.getMethod ("findAll",          null);
      } catch (NoSuchMethodException ne) { LTUtils.handleException (ne, this, null, null);
      }
    }
    protected WriteThroughCache sharedCache () { return (cache == null) ? (cache = new WriteThroughCache (this)) : cache; }
    private static WriteThroughCache cache = null;
}

class CachingAuthorshipHandler extends CachingEntityHandler {
    public CachingAuthorshipHandler (String name) {
      super (name, AuthorshipHome.class);
      try {
        getterModel        = Authorship.    class.getMethod ("getModel",         null);
        setterModel        = Authorship.    class.getMethod ("putModel",         new Class [] { AuthorshipModel.class });
        creatorModel       = AuthorshipHome.class.getMethod ("create",           new Class [] { AuthorshipModel.class });
        finderByPrimaryKey = AuthorshipHome.class.getMethod ("findByPrimaryKey", new Class [] { String.class });
        finderAll          = AuthorshipHome.class.getMethod ("findAll",          null);
        finder1            = AuthorshipHome.class.getMethod ("findByTalkKey",    new Class [] { String.class });
        finder1a           = AuthorshipHome.class.getMethod ("findByAuthorKey",  new Class [] { String.class });
      } catch (NoSuchMethodException ne) { LTUtils.handleException (ne, this, null, null);
      }
    }
    protected WriteThroughCache sharedCache () { return (cache == null) ? (cache = new WriteThroughCache (this)) : cache; }
    private static WriteThroughCache cache = null;
}

class CachingBlockHandler extends CachingEntityHandler {
    public CachingBlockHandler (String name) {
      super (name, BlockHome.class);
      try {
        getterModel        = Block.    class.getMethod ("getModel",         null);
        setterModel        = Block.    class.getMethod ("putModel",         new Class [] { BlockModel.class });
        creatorModel       = BlockHome.class.getMethod ("create",           new Class [] { BlockModel.class });
        finderByPrimaryKey = BlockHome.class.getMethod ("findByPrimaryKey", new Class [] { String.class });
        finderAll          = BlockHome.class.getMethod ("findAll",          null);
        finder1            = BlockHome.class.getMethod ("findByTalkList",   new Class [] { String.class });
      } catch (NoSuchMethodException ne) { LTUtils.handleException (ne, this, null, null);
      }
    }
    protected WriteThroughCache sharedCache () { return (cache == null) ? (cache = new WriteThroughCache (this)) : cache; }
    private static WriteThroughCache cache = null;
}

class CachingPitemHandler extends CachingEntityHandler {
    public CachingPitemHandler (String name) {
      super (name, PitemHome.class);
      try {
        getterModel        = Pitem.    class.getMethod ("getModel",         null);
        setterModel        = Pitem.    class.getMethod ("putModel",         new Class [] { PitemModel.class });
        creatorModel       = PitemHome.class.getMethod ("create",           new Class [] { PitemModel.class });
        finderByPrimaryKey = PitemHome.class.getMethod ("findByPrimaryKey", new Class [] { String.class });
        finderAll          = PitemHome.class.getMethod ("findAll",          null);
        finder1            = PitemHome.class.getMethod ("findByBlockKey",   new Class [] { String.class });
        finder1a           = PitemHome.class.getMethod ("findByTrackKey",   new Class [] { String.class });
      } catch (NoSuchMethodException ne) { LTUtils.handleException (ne, this, null, null);
      }
    }
    protected WriteThroughCache sharedCache () { return (cache == null) ? (cache = new WriteThroughCache (this)) : cache; }
    private static WriteThroughCache cache = null;
}

class CachingTalkHandler extends CachingEntityHandler {
    public CachingTalkHandler (String name) {
      super (name, TalkHome.class);
      try {
        getterModel        = Talk.    class.getMethod ("getModel",         null);
        setterModel        = Talk.    class.getMethod ("putModel",         new Class [] { TalkModel.class });
        creatorModel       = TalkHome.class.getMethod ("create",           new Class [] { TalkModel.class });
        finderByPrimaryKey = TalkHome.class.getMethod ("findByPrimaryKey", new Class [] { String.class });
        finderAll          = TalkHome.class.getMethod ("findAll",          null);
        finder1            = TalkHome.class.getMethod ("findByTypeKey",    new Class [] { String.class });
      } catch (NoSuchMethodException ne) { LTUtils.handleException (ne, this, null, null);
      }
    }
    protected WriteThroughCache sharedCache () { return (cache == null) ? (cache = new WriteThroughCache (this)) : cache; }
    private static WriteThroughCache cache = null;
}

class CachingTrackHandler extends CachingEntityHandler {
    public CachingTrackHandler (String name) {
      super (name, TrackHome.class);
      try {
        getterModel        = Track.    class.getMethod ("getModel",              null);
        setterModel        = Track.    class.getMethod ("putModel",              new Class [] { TrackModel.class });
        creatorModel       = TrackHome.class.getMethod ("create",                new Class [] { TrackModel.class });
        finderByPrimaryKey = TrackHome.class.getMethod ("findByPrimaryKey",      new Class [] { String.class });
        finderAll          = TrackHome.class.getMethod ("findAll",               null);
        finder1            = TrackHome.class.getMethod ("findByPage",            new Class [] { int.class });
        finder1a           = TrackHome.class.getMethod ("findByDate",            new Class [] { java.sql.Date.class });
        finder2            = TrackHome.class.getMethod ("findByPageAndTrackCol", new Class [] { int.class, int.class });
      } catch (NoSuchMethodException ne) { LTUtils.handleException (ne, this, null, null);
      }
    }
    protected WriteThroughCache sharedCache () { return (cache == null) ? (cache = new WriteThroughCache (this)) : cache; }
    private static WriteThroughCache cache = null;
}

class CachingTypeHandler extends CachingEntityHandler {
    public CachingTypeHandler (String name) {
      super (name, TypeHome.class);
      try {
        getterModel        = Type.    class.getMethod ("getModel",         null);
        setterModel        = Type.    class.getMethod ("putModel",         new Class [] { TypeModel.class });
        creatorModel       = TypeHome.class.getMethod ("create",           new Class [] { TypeModel.class });
        finderByPrimaryKey = TypeHome.class.getMethod ("findByPrimaryKey", new Class [] { String.class });
        finderAll          = TypeHome.class.getMethod ("findAll",          null);
      } catch (NoSuchMethodException ne) { LTUtils.handleException (ne, this, null, null);
      }
    }
    protected WriteThroughCache sharedCache () { return (cache == null) ? (cache = new WriteThroughCache (this)) : cache; }
    private static WriteThroughCache cache = null;
}
