/**
 * Title:        Living Talks Web Conference Tool<p>
 * Description:  Tool using J2EE standards<p>
 * Copyright:    Copyright (c) Dr. Falk Langhammer<p>
 * Company:      Living Pages Research GmbH<p>
 * @author Falk Langhammer
 * @version 1.0
 */
package com.livis.livingtalks.ejb;

import com.livis.livingtalks.model.*;
import java.util.*;
import java.lang.reflect.*;
import java.rmi.*;
import javax.ejb.*;

/**
 * EntityHandler is a generic class capable of acting as a fassade to any of our entity beans.
 * In an ideal world, our entity beans would have been derived from a common base class.
 * General algorithms acting on them would then have used the base class interface.
 * We have at least one such algorithm in our system: the cache. Therefore, we need a solution.
 * EJBs are components, not objects, and cannot be derived from each other.<p>
 * The first idea is that their implementations can, but wait: Imagine we would have definded a base method
 * <code>setModel (AbstractModel m)</code> in our entity bean base class. Then, because of RMI-IIOP
 * we would have got a Corba-Exception when calling <code>setModel</code> with, e.g., an <code>AuthorModel m</code>
 * as argument. The solution presented here encapsulates all breaches of object-oriented design in this single file.<p>
 * This is also a nice example where the difference between RMI-IIOP and RMI affects the entire system design.
 */
public class EntityHandler {

    public AbstractModel getModel (String key) {
      if (getterModel == null) throw new LivingTalksException ("Internal Error: No such method: getModel", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          str_arg [0] = key;
          EJBObject entity = (EJBObject) finderByPrimaryKey.invoke (home, str_arg);
          AbstractModel model = (AbstractModel) getterModel.invoke (entity, null);
          return model;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, key); return null;
      } catch (FinderException fe) { return null;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, key); return null;
      }
    }

    public AbstractModel getShortmodel (String key) {
      if (getterModel == null) throw new LivingTalksException ("Internal Error: No such method: getShortmodel", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          str_arg [0] = key;
          EJBObject entity = (EJBObject) finderByPrimaryKey.invoke (home, str_arg);
          AbstractModel model = (AbstractModel) getterModel.invoke (entity, null);
          model.makeShort ();
          return model;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, key); return null;
      } catch (FinderException fe) { return null;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, key); return null;
      }
    }

    public void setModel (AbstractModel model) throws CreateException {
      if (setterModel == null) throw new LivingTalksException ("Internal Error: No such method: setModel", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          str_arg [0] = model.getKey (); model_arg [0] = model;
          EJBObject entity = (EJBObject) finderByPrimaryKey.invoke (home, str_arg);
          setterModel.invoke (entity, model_arg);
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, model);
      } catch (FinderException fe) { throw new LivingTalksCreateException (fe);
      } catch (CreateException ce) { throw ce;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, model);
      }
    }

    public void remove (String key) throws RemoveException {
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        home.remove (key);
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, key);
      } catch (RemoveException de) { throw de;
      }
    }

    public String create (AbstractModel model) throws CreateException {
      if (creatorModel == null) throw new LivingTalksException ("Internal Error: No such method: create", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          model_arg [0] = model;
          EJBObject entity = (EJBObject) creatorModel.invoke (home, model_arg);
          String result = (String) entity.getPrimaryKey ();
          return result;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, model); return null;
      } catch (CreateException ce) { throw ce;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, model); return null;
      }
    }

    public Collection getShortmodelsAll () {
      if (finderAll == null) throw new LivingTalksException ("Internal Error: No such method: ", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          Collection entities = new ArrayList ((Collection) finderAll.invoke (home, null)); // never share
          Collection models   = new ArrayList (entities.size ());
          for (Iterator i = entities.iterator (); i.hasNext (); ) {
            EJBObject entity = (EJBObject) i.next ();
            AbstractModel model = (AbstractModel) getterModel.invoke (entity, null);
            model.makeShort ();
            models.add (model);
          }
          return models;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, null); return null;
      } catch (FinderException fe) { LTUtils.handleException (fe, this, null, null); return null;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, null); return null;
      }
    }

    public Collection getKeysAll () {
      if (finderAll == null) throw new LivingTalksException ("Internal Error: No such method: getKeysAll", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          Collection entities = new ArrayList ((Collection) finderAll.invoke (home, null)); // never share
          Collection keys     = new ArrayList (entities.size ());
          for (Iterator i = entities.iterator (); i.hasNext (); ) {
            EJBObject entity = (EJBObject) i.next ();
            keys.add (entity.getPrimaryKey ());
          }
          return keys;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, null); return null;
      } catch (FinderException fe) { LTUtils.handleException (fe, this, null, null); return null;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, null); return null;
      }
    }

    public Collection getKeys1 (Object a) {
      if (finder1 == null) throw new LivingTalksException ("Internal Error: No such method: getKeys1", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          arg [0] = a;
          Collection entities = new ArrayList ((Collection) finder1.invoke (home, arg)); // never share
          Collection keys     = new ArrayList (entities.size ());
          for (Iterator i = entities.iterator (); i.hasNext (); ) {
            EJBObject entity = (EJBObject) i.next ();
            keys.add (entity.getPrimaryKey ());
          }
          return keys;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, a); return null;
      } catch (FinderException fe) { LTUtils.handleException (fe, this, null, a); return null;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, a); return null;
      }
    }

    public Collection getKeys1a (Object a) {
      if (finder1a == null) throw new LivingTalksException ("Internal Error: No such method: getKeys1a", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          arg [0] = a;
          Collection entities = new ArrayList ((Collection) finder1a.invoke (home, arg)); // never share
          Collection keys     = new ArrayList (entities.size ());
          for (Iterator i = entities.iterator (); i.hasNext (); ) {
            EJBObject entity = (EJBObject) i.next ();
            keys.add (entity.getPrimaryKey ());
          }
          return keys;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, a); return null;
      } catch (FinderException fe) { LTUtils.handleException (fe, this, null, a); return null;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, a); return null;
      }
    }

    public Collection getKeys2 (Object a, Object b) {
      if (finder2 == null) throw new LivingTalksException ("Internal Error: No such method: getKeys2", this);
      try {
        if (home == null) home = LTUtils.lookupHome (homeName, homeClass);
        try {
          args2 [0] = a; args2 [1] = b;
          Collection entities = new ArrayList ((Collection) finder2.invoke (home, args2)); // never share
          Collection keys     = new ArrayList (entities.size ());
          for (Iterator i = entities.iterator (); i.hasNext (); ) {
            EJBObject entity = (EJBObject) i.next ();
            keys.add (entity.getPrimaryKey ());
          }
          return keys;
        } catch (IllegalAccessException ae)    { throw ae;
        } catch (InvocationTargetException ie) { throw ie.getTargetException ();
        }
      } catch (RemoteException re) { LTUtils.handleException (re, this, null, "" + a + "," + b); return null;
      } catch (FinderException fe) { LTUtils.handleException (fe, this, null, "" + a + "," + b); return null;
      } catch (Throwable te)       { LTUtils.handleException (te, this, null, "" + a + "," + b); return null;
      }
    }

    protected Method  getterModel        = null;
    protected Method  setterModel        = null;
    protected Method  creatorModel       = null;
    protected Method  finderByPrimaryKey = null;
    protected Method  finderAll          = null;
    protected Method  finder1            = null;
    protected Method  finder1a           = null;
    protected Method  finder2            = null;

    protected EntityHandler (String homeName, Class homeClass) {
      this.homeName  = homeName;
      this.homeClass = homeClass;
    }

    private   EJBHome home               = null;
    private   String  homeName           = null;
    private   Class   homeClass          = null;

    private String []        str_arg   = new String [1];
    private AbstractModel [] model_arg = new AbstractModel [1];
    private Object []        arg       = new Object [1];
    private Object []        args2     = new Object [2];
}
