/**
 * Title:        Living Talks Web Conference Tool<p>
 * Description:  Tool using J2EE standards<p>
 * Copyright:    Copyright (c) Dr. Falk Langhammer<p>
 * Company:      Living Pages Research GmbH<p>
 * @author Falk Langhammer
 * @version 1.0
 */
package com.livis.livingtalks.ejb;

import java.util.*;
import javax.ejb.CreateException;


/** A thread-safe miss-free write-through cache.
 *  This particular implementation will never forget!
 *  Maybe, one should call it a mirror instead.
 *  @see com.livis.livingtalks.ejb.DataHandler
 */
public class WriteThroughCache implements DataHandler {

  private TreeMap     cache;
  private DataHandler handler;
  private boolean     filled;

  /** Create the cache.
   *  @param handler The cached DataHandler who must do the underlying writes and reads.
   */
  public WriteThroughCache (DataHandler handler) {
    this.cache   = new TreeMap ();
    this.handler = handler;
    this.filled  = false;
  }

  /** @return whether the cache caches all data (do not confuse with exceeded capacity).
   */
  public boolean isFilled () {
    return filled;
  }

  /** @return whether the cache caches all data (do not confuse with exceeded capacity).
   */
  public synchronized void invalidate () {
    cache.clear ();
    setFilled (false);
  }

  /** Set whether the cache has read all data already. Initially, this is false.
   *  @param whether the cache is completely filled.
   */
  public void setFilled (boolean filled) {
    this.filled = filled;
  }

  /** Write new data.
   *  @param data The data to be stored. data must not be <code>null</code>.
   *  @param key  The key under which the data should be stored. key may be <code>null</code>.
   *  @return the key under which the data finally got stored.
   *  @throws CreateException if key does already exist and no other key was used.
   */
  public String writeNew (String key, Object data) throws CreateException {
    if (data == null) throw new CreateException ("New data for key (" + key + ") must not be null");
    String result = handler.writeNew (key, data);
    Object data1  = handler.read     (result);    // handler may have modified data. Attention: reader must see our transaction!
    if (data1 == null) {
      this.invalidate ();
      throw new CreateException ("Warning: data for key (" + result + ") was written but could not be cached - invalidating cache");
    }
    synchronized (this) { cache.put (result, data1); }
    return result;
  }

  /** Updates old data.
   *  @param data the data to be stored. A value of <code>null</code> will delete the data.
   *  @param key  The key under which to store the data. key must not be <code>null</code>.
   *  @throws CreateException if key does not exist
   */
  public void writeOld (String key, Object data) throws CreateException {
    if (key == null) throw new CreateException ("Could not update old data because no key was given");
    handler.writeOld (key, data);
    if (data == null) synchronized (this) {
      cache.remove (key);
    }
    else {
      Object data1 = handler.read (key);
      if (data1 != null) synchronized (this) { cache.put (key, data1); }
    }
  }

  /** Read data stored at key.
   *  @param key The key to retrieve the data for. A value of <code>null</code> will return null.
   *  @return The object previously stored, or <code>null</code> if no such data exists.
   *  @throws CreateException if key does not exist
   */
  public Object read (String key) {
    if (key == null) return null;
    Object data = null;
    synchronized (this) { data = cache.get (key); }
    if (isFilled () || data != null) return data;
    data = handler.read (key);   // (handler may call us back)
    synchronized (this) { if (data != null) cache.put (key, data); }
    return data;
  }

  /** @return all keys sorted into a Collection if cache is filled, null otherwise.
   */
  public synchronized Collection getAllKeys () {
    if (isFilled ()) return new ArrayList (cache.keySet ());
    else             return null;
  }

  /** @return all data as a collection (no entry is null) if cache is filled, null otherwise.
   */
  public synchronized Collection getAllData () {
    if (isFilled ()) return new ArrayList (cache.values ());
    else             return null;
  }
}