/**
 * Title:        Living Talks Web Conference Tool<p>
 * Description:  Tool using J2EE standards<p>
 * Copyright:    Copyright (c) Dr. Falk Langhammer<p>
 * Company:      Living Pages Research GmbH<p>
 * @author Falk Langhammer
 * @version 1.0
 */
package com.livis.livingtalks.model;

import com.livis.flabes.util.*;
import com.livis.flabes.util.ejb.*;
import java.io.*;
import java.util.*;
import java.text.SimpleDateFormat;

abstract public class AbstractModel
    implements Serializable {

    //----------------------------------------------------------------------
    // constructors/assignment
    //----------------------------------------------------------------------
    protected AbstractModel assignAbstractModel (String key, String editor, String timestamp) {
	setKey       (key);
	setEditor    (editor);
	setTimestamp (timestamp);

	return this;
    }

    /**
     * Modifies the model such that less relevant properties
     * are deleted from the model.
     * The version of AbstractModel does nothing. A derived version
     * may set a memory-consuming property to null if it is not used in any
     * master or tabular view, i.e. a view of many models at once.
     * A derived version may set a flag if makeShort got called and throw a
     * LivingTalksException if a property is gotten which was nulled by makeShort.
     **/
    public void makeShort () {
    }

    /**
     * Set this Model to a clear state suitable for data entry into a
     * clear form it is not recommended behaviour for a clear model to
     * pass a subsequent {@link #check} test.
     **/
    public void clear () { setKey (isKeyCreatable () ? getAutoKey () : ""); touch (); }

    //----------------------------------------------------------------------
    // CMP fields/properties
    //----------------------------------------------------------------------
    public String key;
    public String getKey       ()                 { return key; }
    public void   setKey       (String key)       { this.key = key; }

    public String editor;
    public String getEditor    ()                 { return editor; }
    public void   setEditor    (String editor)    { this.editor = editor; }

    public String timestamp;
    public String getTimestamp ()                 { return timestamp; }
    public void   setTimestamp (String timestamp) { this.timestamp = timestamp; }

    // we do not call this "setTimestamp(Date)" because the Sun's bean property finding framework then misses timestamp
    public void   setTimestampFromDate (Date date){ setTimestamp ((new SimpleDateFormat ("dd.MM.yyyy HH:mm")).format (date)); }
    public void   touch        ()                 { setTimestampFromDate (new Date ()); }

    //----------------------------------------------------------------------
    // key creation/check
    //----------------------------------------------------------------------
    /**
     * Checks whether a key can be created from other model data.
     *
     * This method is called by {@link #check} if {@link
     * #isKeyCreatable} returns true. This default implementation does
     * nothing and should be re-implemented by derived classes in
     * conjunction with {@link #isKeyCreatable} and {@link
     * #createKey}.
     *
     * @throws LivingTalksException if a key cannot be created from
     *                              other model data.
     *
     * @see    #isKeyCreatable
     * @see    #createKey
     * @see    #check
     **/
    public void checkForKeyCreation () { }

    /**
     * Checks if model is logically OK.
     *
     * This default implementation checks for an non-empty key and --
     * if {@link #isKeyCreatable} returns true -- whether the key can
     * in fact be created (using {@link #checkForKeyCreation}).
     *
     * Also, special charcters (<, >, &, ", ', \r, and \n) are removed
     * from the key.
     *
     * @throws LivingTalksException if model is invalid. Must not
     *                              check for duplicate keys as this
     *                              is checked anyway at database
     *                              insertion time.
     * @see    #isKeyCreatable
     * @see    #createKey
     **/
    public void check () {
	if (isKeyCreatable ()) {
	    checkForKeyCreation ();
	}
	checkForNonEmptyKey ();
    }

    private String autoKey = null;
    /**
     * A text instructing the user about what are legal key
     * values.
     *
     * When left unchanged by user it will trigger a {@link
     * #createKey} method. Therefore, the text should state also
     * whether this is ok or not.
     **/
    public String getAutoKey ()               {
	if (null == autoKey) {
	    // not set explicitly, return standard text
	    if (isKeyCreatable ()) {
		return "[don't edit for automatic key generation]";
	    } else {
		return "[replace by a unique key value]";
	    }
	} else {
	    return autoKey;
	}
    }
    public void   setAutoKey (String autoKey) { this.autoKey = autoKey; }

    /**
     * Checks whether a key can, in theory, be created from other
     * model data.
     *
     * This default implementation returns <code>false</code>. If
     * derived classes re-implement this to return <code>true</true>,
     * they should also overwrite {@link #createKey} to return a
     * suitable key and {@link #checkForKeyCreation} to test whether a
     * key is valid.
     *
     * @return  <code>true</code> if a key can be created
     * @see     #createKey
     * @see     #checkForKeyCreation
     **/
    public boolean isKeyCreatable () { return false; }

    /**
     * Creates a key from other model data which must pass the {@link
     * #check} test if {@link #isKeyCreatable} returns
     * <code>true</code>.
     *
     * The generated key is returned. This default implementation
     * always throws an exception.
     *
     * @returns the generated key.
     * @throws  LivingTalksException if cannot create a key.
     * @see     #isKeyCreatable
     * @see     #check
     **/
    protected String createKey () {
	throw new LivingTalksException ("I cannot create the key - You will have to provide the key value!", this);
    }

    protected final boolean shallCreateKey () {
	return isKeyCreatable () && getAutoKey ().equals (getKey ());
    }

    private static char[] specialChars = new char[] { '<', '>', '&', '"', '\'', '\r', '\n', };
    /**
     * Checks if a key value is not empty.
     *
     * @throws LivingTalksException if key is empty.
     **/
    protected final void checkForNonEmptyKey () {
	if (StringUtils.isNullOrEmpty (getKey ()) || getAutoKey ().equals (getKey ())) {
	    throw new LivingTalksException ("The key is mandatory!", this);
	}
	String oldKey = new String (getKey ());
	setKey (StringUtils.removeOccurrences (getKey (), specialChars));
	if (StringUtils.isNullOrEmpty (getKey ())) {
	    throw new LivingTalksException ("After removal of special characters <, >, &, \", ',"
					    + " <carriage return>, and <line feed> the key is empty (was: " + oldKey + ")!", this);
	}

    }

    //----------------------------------------------------------------------
    // XML
    //----------------------------------------------------------------------
    protected abstract List xmlTagPropertiesStore ();
    private List xmlTagProperties;
    public List XMLTagProperties () {
	if (null == xmlTagProperties) {
	    if (xmlTagPropertiesStore ().isEmpty ()) {
		ClassUtils.assureDerivedClassHasDefined (getClass (), "xmlTagPropertiesStore");
		xmlTagPropertiesStore ().addAll (EJBUtils.getCMPfieldNames (getClass ()));
	    }
	    xmlTagProperties = xmlTagPropertiesStore ();
	}
	return xmlTagProperties;
    }
}
