/**
 * Title:        Living Talks Web Conference Tool<p>
 * Description:  Tool using J2EE standards<p>
 * Copyright:    Copyright (c) Dr. Falk Langhammer<p>
 * Company:      Living Pages Research GmbH<p>
 * @author Falk Langhammer
 * @version 1.0
 */
package com.livis.livingtalks.model;

import com.livis.flabes.util.*;
import com.livis.flabes.util.ejb.*;
import com.livis.livingtalks.ejb.*;
import java.lang.reflect.*;
import java.rmi.*;
import java.util.*;
import javax.ejb.*;
import javax.naming.*;
import javax.rmi.*;
import javax.transaction.RollbackException;
import org.xml.sax.SAXException;
import org.omg.CORBA.OBJECT_NOT_EXIST;

public class LTUtils {

    private static final String remote   = "The application server did complain!<br>The complaint is: ";
    private static final String noobject = "A remote server object disappeared.<br>\nNormally, this is a follow-up error to a previous problem.<br>\nYou should <b> logout and login </b> again now!";
    private static final String sax      = "The XML processing failed!<br>Was processing: ";
    private static final String name     = "The application server could not be contacted!<br>Was looking for: ";
    private static final String duplic   = "The entry does already exist!<br>&nbsp;<br>You must choose another key or name.<br>It was: ";
    private static final String create   = "The information could not be added!<br>Was trying to add: ";
    private static final String find     = "The information could not be found!<br>Was searching for: ";
    private static final String delete   = "The information could not be deleted!<br>Was trying to delete: ";
    private static final String rollback = "The database did ignore your request!<br>Possibly, some field contained too many characters?<br>(255 and 4000 are typical limits.)<br>The item was: ";
    private static final String nan      = "You are supposed to enter a number!<br>This is not a number: ";
    private static final String unknown  = "There is a problem!<br>Was processing: ";

    private LTUtils () { } // no instances needed

    public static void handleException (Throwable ex, Object thrower, String message, Object item) {
	if (ex != null) ex.printStackTrace ();
        // We want to handle the innermost, real cause of the problem - except if it is our own LivingTalksException
	Throwable nex = ex, nested = null;
	while (nex instanceof RemoteException            && (nested = ((RemoteException)            nex).detail)                != null
            || nex instanceof InvocationTargetException  && (nested = ((InvocationTargetException)  nex).getTargetException ()) != null
	    || nex instanceof SAXException               && (nested = ((SAXException)               nex).getException ())       != null
	    || nex instanceof LivingTalksCreateException && (nested = ((LivingTalksCreateException) nex).getNested ())          != null)
        {
          nex = nested;
        }
	if (nex instanceof LivingTalksException)
	    throw (LivingTalksException) nex;

	String text;
	if      (message != null)                      { text = message; }
	else if (nex instanceof NoSuchObjectException) { text = noobject; }
	else if (nex instanceof OBJECT_NOT_EXIST)      { text = noobject; }
	else if (nex instanceof RemoteException)       { text = remote + nex.getLocalizedMessage (); }
	else if (nex instanceof SAXException)          { text = sax    + nex.getLocalizedMessage (); }
	else if (nex instanceof NamingException)       { text = name; }
	else if (nex instanceof DuplicateKeyException) { text = duplic; }
	else if (nex instanceof CreateException)       { text = create; }
	else if (nex instanceof FinderException)       { text = find; }
	else if (nex instanceof RemoveException)       { text = delete; }
	else if (nex instanceof RollbackException)     { text = rollback; }
	else if (nex instanceof NumberFormatException) { text = nan; }
	else                                           { text = unknown; }
	if (item != null) text += " \"" + item + "\"";
	text += "\n";

	if (nex != null && nex != ex) throw new LivingTalksException (text, thrower, ex, nex);
	else                          throw new LivingTalksException (text, thrower, ex);
    }

    public static int parseInt (String s) throws LivingTalksException {
	try { return Integer.parseInt (s); }
        catch (NumberFormatException e) { handleException (e, null, null, s); return 0; }
    }

    //----------------------------------------------------------------------
    private static String openParagraph  = "<p class=\"ap\">";
    private static String closeParagraph = "</p>";
    private static String lineBreak      = "<br />";
    public static String preformattedToHTML (String s) {
	if (StringUtils.isNullOrEmpty (s)) {
	    return s;
	}

	// DOS
	String result = StringUtils.replace (s, "\r\n", "\n");

	result = StringUtils.cutLeading  (result, '\n');
	result = StringUtils.cutTrailing (result, '\n');

	result = StringUtils.replace (result, "\n\n",                         closeParagraph + openParagraph);
	result = StringUtils.replace (result, "\n",                           lineBreak + "\n");
	result = StringUtils.replace (result, closeParagraph + openParagraph, closeParagraph + "\n" + openParagraph);

	return openParagraph + result + closeParagraph;
    }

    //----------------------------------------------------------------------
    public static String createKey (String s, int minWordLength, int maxKeyLength) {
	String[] words        = StringUtils.split (s, ' ');
	int      firstLongWord;
	boolean  found        = false;
	for (firstLongWord = 0; firstLongWord < words.length; ++firstLongWord) {
	    // yes, this is an assignment
	    if ((found = (minWordLength <= words [firstLongWord].length ()))) {
		break;
	    }
	}
	String key = StringUtils.concat (words, "_", found ? firstLongWord : 0);
	return key.substring (0, Math.min (maxKeyLength, key.length ())).trim ();
    }

    public static String uniqueKey (Manage mgr, String getModelMethod, String keyPrefix) {
	Method getter = ClassUtils.getMethodCatching (mgr.getClass (), getModelMethod, new Class [] { String.class });
	if (getter == null) throw new LivingTalksException ("Internal Error: no such method: Manage." + getModelMethod, LTUtils.class);
	for (int nTry = 0; /* loop until null is returned */; ++nTry) {
	    // in first try (#0) use key as is, otherwise append the trial number:
	    String key = keyPrefix + ((0 < nTry) ? ("." + nTry) : "");
	    // if we do not find a model for this key, then it may be used to store a new model:
	    try {
              if (getter.invoke (mgr, new Object [] { key }) == null) return key;
            } catch (InvocationTargetException ie) { handleException (((InvocationTargetException) ie).getTargetException (), LTUtils.class, null, keyPrefix);
            } catch (IllegalAccessException ae)    { handleException (ae,                                                     LTUtils.class, null, keyPrefix);
            }
	}
    }

    //----------------------------------------------------------------------
    // naming context/homes
    //----------------------------------------------------------------------
    public static final String Manage_BEAN     = "ejb/manage";
    public static final String Publish_BEAN    = "ejb/publish";

    private static Context     context;
    static {
	context = null;
    }

    public static EJBHome lookupHome (String name, Class cls) {
    	try {
	    if (context == null) context = new InitialContext ();
	    return (EJBHome) PortableRemoteObject.narrow (context.lookup (name), cls);
	} catch (NamingException e) {
	    LTUtils.handleException (e, null, null, name);
	    return null;
	}
    }

    private static ManageHome manageHome;
    public  static ManageHome getManageHome () {
	return (null ==  manageHome) ?
	    (manageHome     = (ManageHome)     lookupHome (Manage_BEAN,     ManageHome.class))     : manageHome;
    }

    public  static Manage createManage ()      {
    	try {
            return getManageHome ().create ();
	} catch (CreateException ce) { LTUtils.handleException (ce, LTUtils.class, "Internal Error: No manager available!", null); return null;
	} catch (RemoteException re) { LTUtils.handleException (re, LTUtils.class, "Internal Error: No manager available!", null); return null;
	}
    }

    private static PublishHome publishHome;
    public  static PublishHome getPublishHome () {
	return (null ==  publishHome) ?
	    (publishHome    = (PublishHome)    lookupHome (Publish_BEAN,    PublishHome.class))    : publishHome;
    }
}
