/**
 * Title:        Living Talks Web Conference Tool<p>
 * Description:  Tool using J2EE standards<p>
 * Copyright:    Copyright (c) Dr. Falk Langhammer<p>
 * Company:      Living Pages Research GmbH<p>
 * @author Falk Langhammer
 * @version 1.0
 */
package com.livis.livingtalks.xml;

import com.livis.livingtalks.model.LivingTalksException;
import com.livis.flabes.util.Indenter;
import java.util.*;
import java.io.PrintWriter;


/** A utility object which computes a 2D layout of tablecells which will hold all blocks.
 *  The output of this map is used to help a XML style shhet to create a HTML table layout.
 */
public class Tablemap {

  class Item {
    String key;
    int    colStart;
    int    colLength;
    int    rowStart;
    int    rowLength;
    Item (String key, int colStart, int colLength, int rowStart, int rowLength) {
      this.key        = key;
      this.colStart   = colStart;
      this.colLength  = colLength;
      this.rowStart   = rowStart;
      this.rowLength  = rowLength;
    }
  }

  class Cell {
    boolean empty    = true;
    boolean covered  = false;
    int     cols     = 1;
    int     rows     = 1;
    Item    item     = null;
  }

  private int        page;
  private Collection items;

  public Tablemap (int page) {
    this.page  = page;
    this.items = new ArrayList ();
  }

  public void add (String key, int page, int colStart, int colLength, int rowStart, int rowLength) {
    if (page != this.page) throw new LivingTalksException ("Layout Failure: Tablemap encountered inconsistency: attempt to map entry " + key + " of page " + page + " into page " + this.page, this);
    items.add (new Item (key, colStart, colLength, rowStart, rowLength));
  }

  public void clear () {
    items.clear ();
  }

  public void print (PrintWriter out, Indenter indenter, String tag, int colRaster, int colStart, int colLength, int rowRaster, int rowStart, int rowLength) {
    // create array of cells:
    int cols = colLength / colRaster;
    int rows = rowLength / rowRaster;
    Cell [][] table = new Cell [cols][rows];
    for (int c = 0; c < cols; ++c) for (int r = 0; r < rows; ++r) table [c][r] = new Cell ();

    // put items into table:
    for (Iterator i = items.iterator (); i.hasNext (); ) {
      Item item  = (Item) i.next ();
      int  icol  = (item.colStart - colStart)  / colRaster;
      int  icols = item.colLength / colRaster;
      int  irow  = (item.rowStart - rowStart)  / rowRaster;
      int  irows = item.rowLength / rowRaster;

      if (icols < 1 || irows < 1) {
        throw new LivingTalksException ("Layout Error: Tablemap discovered that entry " + item.key + " is too small to print - ignored!", this);
        // continue;
      }
      if (icol < 0 || icol + icols > cols || irow < 0 || irow + irows > rows) {
        throw new LivingTalksException ("Layout Error: Tablemap discovered that entry " + item.key + " is off page - ignored!", this);
        // continue;
      }
      boolean empty = true;
      for (int ic = icol; ic < icol + icols; ++ic) for (int ir = irow; ir < irow + irows; ++ir) empty &= table [ic][ir].empty;
      if (!empty) {
        throw new LivingTalksException ("Layout Error: Tablemap discovered that entry " + item.key + " is overlapping another - ignored!", this);
        // continue;
      }
      for (int ic = icol; ic < icol + icols; ++ic) for (int ir = irow; ir < irow + irows; ++ir) {
        table [ic][ir].empty   = false;
        table [ic][ir].covered = true;
      }
      table [icol][irow].covered = false;
      table [icol][irow].cols    = icols;
      table [icol][irow].rows    = irows;
      table [icol][irow].item    = item;
    }

    // collapse empty space:
    for (int c = 0; c < cols; ++c) {
      boolean was_empty = false;
      int     start_r   = 0;
      for (int r = 0; r < rows; ++r) {
        if (!was_empty && table [c][r].empty) {
          start_r = r;
          was_empty = true;
        }
        else if (was_empty && !table [c][r].empty) {
          for (int ir = start_r; ir < r; ++ir) table [c][ir].covered = true;
          table [c][start_r].covered = false;
          table [c][start_r].rows    = r - start_r;
          was_empty = false;
        }
        else if (was_empty && (r + 1) == rows) {
          for (int ir = start_r; ir < (r + 1); ++ir) table [c][ir].covered = true;
          table [c][start_r].covered = false;
          table [c][start_r].rows    = (r + 1) - start_r;
        }
      }
    }

    // write output:
    out.println (indenter + "<tablemap page=\"" + page + "\" ncols=\"" + cols + "\" nrows=\"" + rows + "\">");
    out.println (indenter.incLevel () + "<thr>");
    indenter.incLevel ();
    for (int c = 0; c < cols; ++c) {
      out.println (indenter + "<thd col=\"" + c + "\">");
      int col = colStart + c * colRaster;
      out.println (indenter.incLevel () + "<col-ref>" + col + "</col-ref>");
      out.println (indenter.decLevel () + "</thd>");
    }
    out.println (indenter.decLevel () + "</thr>");
    for (int r = 0; r < rows; ++r) {
      out.println (indenter + "<tr row=\"" + r + "\">");
      indenter.incLevel ();
      for (int c = 0; c < cols; ++c) {
        if (!table [c][r].covered) {
          out.print (indenter + "<td col=\"" + c + "\"");
            if (table [c][r].cols != 1) out.print (" colspan=\"" + table [c][r].cols + "\"");
            if (table [c][r].rows != 1) out.print (" rowspan=\"" + table [c][r].rows + "\"");
          if (table [c][r].empty) {
            out.println ("/>");
          }
          else {
            out.println (">");
            indenter.incLevel ();
            if (table [c][r].item != null) {
              String key = (tag == null) ? "key" : tag;
              out.println (indenter + "<"+tag+">" + ((table [c][r].item.key == null) ? "" : table [c][r].item.key) + "</"+tag+">");
              out.println (indenter + "<colStart>"  + table [c][r].item.colStart  + "</colStart>");
              out.println (indenter + "<colLength>" + table [c][r].item.colLength + "</colLength>");
              out.println (indenter + "<rowStart>"  + table [c][r].item.rowStart  + "</rowStart>");
              out.println (indenter + "<rowLength>" + table [c][r].item.rowLength + "</rowLength>");
            }
            out.println (indenter.decLevel () + "</td>");
          }
        }
      }
      out.println (indenter.decLevel () + "</tr>");
    }
    out.println (indenter.decLevel () + "</tablemap>");
  }
}
