package com.livis.flabes.util;

import java.io.*;
import java.lang.reflect.*;

public class MethodUtils {

    private MethodUtils () { }

    /**
     * Invoke a method by calling {@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * The stack trace of a caught exception is printed.
     *
     * @param   object object to invoke the method on. If the method
     *        	       is static, <code>object</code> is ignord and
     *        	       may be <code>null</code>.
     * @param 	method method to invoke.
     * @param 	args   actual parameters.
     * @param   out    destination to print a caught exception to. No
     *                 exceptions are printed if <code>out</code> is
     *                 <code>null</code>.
     * @returns        result of the method invocation, <code>null</code>
     *                 if an exception was caught.
     * @see            java.lang.reflect.Method#invoke
     **/
    public static Object invokeCatching (Object object, Method method, Object[] args, PrintWriter out) {
	try {
	    return method.invoke (object, args);
	} catch (Exception e) {
	    if (null != out) {
		e.printStackTrace (out);
		out.flush ();
	    }
	    return null;
	}
    }

    private static PrintWriter defaultOut = new PrintWriter (System.err);

    /**
     * Invoke a method by calling {@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Object,Method,Object[],PrintWriter}. The stack
     * trace of a caught exception is printed to {@link
     * #java.lang.System.err}.
     **/
    public static Object invokeCatching (Object object, Method method, Object[] args) {
	return invokeCatching (object, method, args, defaultOut);
    }

    /**
     * Invoke a method by calling {@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Object,Method,Object[],PrintWriter}. The method
     * must have an empty parameter list.
     **/
    public static Object invokeCatching (Object object, Method method, PrintWriter out) {
	return invokeCatching (object, method, new Object [0], out);
    }

    /**
     * Invoke a method by calling {@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Object,Method,Object[],PrintWriter}. The method
     * must have an empty parameter list. The stack trace of a caught
     * exception is printed to {@link #java.lang.System.err}.
     **/
    public static Object invokeCatching (Object object, Method method) {
	return invokeCatching (object, method, new Object [0], defaultOut);
    }

    //----------------------------------------------------------------------
    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * The stack trace of a caught exception is printed.
     *
     * @param   cls        containing class of the method to invoke.
     * @param   object     object to invoke the method on. If the method
     *        	       	   is static, <code>object</code> is ignored
     *        	       	   and may be <code>null</code>.
     * @param 	methodName method to invoke.
     * @param   paramTypes types of the method parameters.
     * @param 	args   	   actual parameters.
     * @param   out    	   destination to print a caught exception to. No
     *                 	   exceptions are printed if <code>out</code> is
     *                 	   <code>null</code>.
     * @returns            result of the method invocation, <code>null</code>
     *                     if an exception was caught.
     * @see                java.lang.reflect.Method#invoke
     * @see                java.lang.Class#getMethod
     **/
    public static Object invokeCatching (Class cls, Object object, String methodName, Class[] paramTypes, Object[] args, PrintWriter out) {
	Method method = ClassUtils.getMethodCatching (cls, methodName, paramTypes, out);
	return invokeCatching (object, method, args, out);
    }

    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Class,Object,String,Class[],Object[],PrintWriter)}. The
     * stack trace of a caught exception is printed to {@link
     * #java.lang.System.err}.
     **/
    public static Object invokeCatching (Class cls, Object object, String methodName, Class[] paramTypes, Object[] args) {
	return invokeCatching (cls, object, methodName, paramTypes, args, defaultOut);
    }

    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Class,Object,String,Class[],Object[],PrintWriter)}. The
     * method must have an empty parameter list.
     **/
    public static Object invokeCatching (Class cls, Object object, String methodName,                                    PrintWriter out) {
	return invokeCatching (cls, object, methodName, new Class [0], new Object [0], out);
    }

    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Class,Object,String,Class[],Object[],PrintWriter)}. The
     * method must have an empty parameter list. The stack trace of a
     * caught exception is printed to {@link #java.lang.System.err}.
     **/
    public static Object invokeCatching (Class cls, Object object, String methodName) {
	return invokeCatching (cls, object, methodName, new Class [0], new Object [0], defaultOut);
    }

    //----------------------------------------------------------------------
    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * The containing class of the method to invoke is extracted from
     * <code>object</code>, that is, even for static methods an Object
     * must be provided.
     *
     * The stack trace of a caught exception is printed.
     *
     * @param   object     object to invoke the method on.
     * @param 	methodName method to invoke.
     * @param   paramTypes types of the method parameters.
     * @param 	args   	   actual parameters.
     * @param   out    	   destination to print a caught exception to. No
     *                 	   exceptions are printed if <code>out</code> is
     *                 	   <code>null</code>.
     * @returns            result of the method invocation, <code>null</code>
     *                     if an exception was caught.
     * @see                java.lang.reflect.Method#invoke
     * @see                java.lang.Class#getMethod
     **/
    public static Object invokeCatching (Object object, String methodName, Class[] paramTypes, Object[] args, PrintWriter out) {
	return invokeCatching (object.getClass (), object, methodName, paramTypes, args, out);
    }

    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Object,String,Class[],Object[],PrintWriter)}. The
     * stack trace of a caught exception is printed to {@link
     * #java.lang.System.err}.
     **/
    public static Object invokeCatching (Object object, String methodName, Class[] paramTypes, Object[] args) {
	return invokeCatching (object, methodName, paramTypes, args, defaultOut);
    }

    // this works for non-null parameters only
    private static Class[] paramTypes (Object[] args) {
	Class[] result = new Class [args.length];
	for (int i = 0; i < args.length; ++i) {
	    result [i] = args [i].getClass ();
	}
	return result;
    }

    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Object,String,Class[],Object[],PrintWriter)}. The
     * stack trace of a caught exception is printed to {@link
     * #java.lang.System.err}. The parameter types are identified from
     * the parameter objects, thus none of these must be
     * <code>null</code>.
     **/
    public static Object invokeCatching (Object object, String methodName,                     Object[] args) {
	return invokeCatching (object, methodName, paramTypes (args), args);
    }

    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Object,String,Class[],Object[],PrintWriter)}. The
     * method must have an empty parameter list.
     **/
    public static Object invokeCatching (Object object, String methodName,                                    PrintWriter out) {
	return invokeCatching (object, methodName, new Class [0], new Object [0], out);
    }

    /**
     * Get and invoke a method by calling {@link
     * java.lang.Class#getMethod}/{@link
     * java.lang.reflect.Method#invoke} and catch all exceptions.
     *
     * Like {@link
     * #invokeCatching(Object,String,Class[],Object[],PrintWriter)}. The
     * method must have an empty parameter list. The stack trace of a
     * caught exception is printed to {@link #java.lang.System.err}.
     **/
    public static Object invokeCatching (Object object, String methodName) {
	return invokeCatching (object, methodName, new Class [0], new Object [0], defaultOut);
    }

    /**
     * Serialize a {@link java.lang.reflect.Method} object.
     **/
    public static void write (ObjectOutputStream out, Method method) throws IOException {
	out.writeObject (method.getDeclaringClass ());
	out.writeObject (method.getName ());
	out.writeObject (method.getParameterTypes ());
    }

    /**
     * De-serialize a {@link java.lang.reflect.Method} object.
     **/
    public static Method read (ObjectInputStream in) throws ClassNotFoundException, IOException {
	try {
	    return ((Class)in.readObject ()).getMethod ((String)in.readObject (), (Class[])in.readObject ());
	} catch (NoSuchMethodException e) {
	    e.printStackTrace (System.err);
	    return null;
	}
    }
}
