package com.livis.flabes.util;

import java.util.*;
import java.net.*;

public class URLUtils {

    private URLUtils () { }

    // file:bla -> bla
    public static String trimProtocol (String urlSpec) {
	int colonPos = urlSpec.indexOf (":");
	if (0 < colonPos) {
	    return urlSpec.substring (colonPos + 1);
	} else {
	    return urlSpec;
	}
    }

    // jar:file:bla -> bla
    public static String trimProtocols (String urlSpec) {
	int colonPos = urlSpec.lastIndexOf (":");
	if (0 < colonPos) {
	    return urlSpec.substring (colonPos + 1);
	} else {
	    return urlSpec;
	}
    }

    /**
     * Check whether a given {@link java.net.URL URL} specification is
     * relative.
     *
     * All URLs which are not well-formed are considered relative.
     *
     * @param   urlSpec string reperesentation of the URL to check
     * @returns         <code>true</code> if <code>urlSpec</code> is
     *                  relative (ie, not well-formed),
     *                  <code>false</code> otherwise
     **/
    public static boolean isRelative (String urlSpec) {
	// we have to check a String and not a URL since you won't be
	// able to construct a non-absolute URL object
	try {
	    new URL (urlSpec);
	} catch (MalformedURLException e) {
	    return true;
	}
	return false;
    }

    /**
     * Create string representation of a relative URL specifying the
     * path between two locations.
     *
     * <code>fromSpec</code> and <code>toSpec</code> are expected to
     * be relative to the same context, which should be omitted. If
     * one or both specifications are given absolute but relative to a
     * context URL specified in <code>contextSpecs</code>, the context
     * prefix with the longest match is removed before contructing the
     * resulting link. Prefixes not mentioned in
     * <code>contextSpecs</code> but common to <code>fromSpec</code>
     * and <code>toSpec</code> are also removed.
     *
     * @param fromSpec     string representation of the "from" location
     * @param toSpec       string representation of the "to" location
     * @param contextSpecs string representations of locations "from"
     *                     and "to " may be relative to
     * @returns            string representations of the link from
     *                     <code>fromSpec</code> to <code>toSpec</code>

     * @throws MalformedURLException if one of the
     *                               <code>contextSpecs</code> is not
     *                               well-formed (<code>null</code>
     *                               references are allowed for
     *                               convenience)
     *
     **/
    public static String relativeURLSpec (String fromSpec, String toSpec, String[] contextSpecs) throws MalformedURLException {
	if (debug) {
	    System.err.println ("from:     " + fromSpec);
	    System.err.println ("to:       " + toSpec);
	    System.err.println ("contexts: " + ((null == contextSpecs) ? null : Arrays.asList (contextSpecs)));
	}

	if ((null == fromSpec) || (null == toSpec)) {
	    if (debug) System.err.println ("result:   " + toSpec);
	    return toSpec;
	}

	String shortFromSpec = fromSpec;
	String shortToSpec   = toSpec;

	// if fromSpec or toSpec is absolute, try to make them
	// relative by removing the longest contextSpec prefix the
	// spec may start with
	if ((!isRelative (fromSpec) || !isRelative (toSpec)) && (null != contextSpecs)) {
	    for (int i = 0; i < contextSpecs.length; ++i) {
		if (null != contextSpecs [i]) {
		    // normalize contextSpecs (trailing slash and the
		    // like) using the URL constructor; this may throw
		    // a MalformedURLExcpetion which is intended
		    // (context URLs must be absolute and well-formed)
		    String normalizedPrefix = new URL (contextSpecs [i]).toString ();
		    
		    if (fromSpec.startsWith (normalizedPrefix)) {
			shortFromSpec = StringUtils.shortest (shortFromSpec, fromSpec.substring (normalizedPrefix.length ()));
		    }
		    if (toSpec.startsWith (normalizedPrefix)) {
			shortToSpec   = StringUtils.shortest (shortToSpec,   toSpec  .substring (normalizedPrefix.length ()));
		    }
		}
	    }
	}

	fromSpec = shortFromSpec;
	toSpec   = shortToSpec;

	try {
	    // if toSpec is still absolute it doesn't start with a
	    // known context prefix; to make it relative, fromSpec
	    // must also be absolute (no known context prefix removed
	    // yet) and only different in the file part (that is,
	    // protocol, host, port etc must be equal); if fromSpec is
	    // not relative, the URL constructor will throw a
	    // MalformedURLExcpetion which is caught below and toSpec
	    // is returnrd unchanged
	    if (!isRelative (toSpec)) {
		URL    fromURL     = new URL (fromSpec);
		URL    toURL       = new URL (toSpec);
		String fromString  = fromURL.toString ();
		String toString    = toURL.toString ();
		fromSpec = fromURL.getFile ();
		toSpec   = toURL  .getFile ();
		String fromNonFile = fromString.substring (0, fromString.length () - fromSpec.length ());
		String toNonFile   = toString  .substring (0, toString  .length () - toSpec  .length ());
		if (!fromNonFile.equals (toNonFile)) {
		    // not the same prefix, can't make it relative
		    if (debug) System.err.println ("result:   " + toSpec);
		    return toSpec;
		}
	    }

	    // now both specs are relative (either known or
	    // unknown-but-equal context prefixes removed); split into
	    // path components and remove common parts
	    String[] fromPath = StringUtils.split (fromSpec, '/', false);
	    String[] toPath   = StringUtils.split (toSpec,   '/', false);

	    String[][] cutPaths    = (String[][])ArraysUtils.cutCommonPrefix (new String[][] { fromPath, toPath, });
	    String[]   cutFromPath = cutPaths [0];
	    String[]   cutToPath   = cutPaths [1];

	    if (0 == cutFromPath.length) {
		// even last (non-directory) part removed, that is,
		// from is prefix of or equal to to

 		if (0 == cutToPath.length) {
		    // from and to are equal

		    if (0 == toPath.length) {
			// toSpec was "", return it (FIXME: what does
			// this mean?)
			if (debug) System.err.println ("result:   " + toSpec);
			return toSpec;
		    } else {
			// let to point itself
			if (debug) System.err.println ("result:   " + toPath [toPath.length - 1]);
			return toPath [toPath.length - 1];
		    }
		} else {
		    // reconstruct toSpec from what is left
		    if (debug) System.err.println ("result:   " + StringUtils.concat (cutToPath, "/"));
		    return StringUtils.concat (cutToPath, "/");
		}
	    } else {
		if (debug) System.err.println ("result:   " + StringUtils.stringOf ("../", cutFromPath.length - 1) + StringUtils.concat (cutToPath, "/"));
		return StringUtils.stringOf ("../", cutFromPath.length - 1) + StringUtils.concat (cutToPath, "/");
	    }
	} catch (MalformedURLException e) {
	    // leave as is
	    if (debug) System.err.println ("result:   " + toSpec);
	    return toSpec;
	}
    }

    //----------------------------------------------------------------------
    private static boolean debug = "true".equals (System.getProperty (URLUtils.class.getName () + ".debug"));
}
