package com.livis.flabes.util.beans;

import com.livis.flabes.util.*;
import java.awt.Image;
import java.beans.*;
import java.util.*;

/**
 * Utility {@link java.beans.BeanInfo BeanInfo} base class to replace
 * {@link java.beans.SimpleBeanInfo SimpleBeanInfo}.
 *
 * It serves as base class for {@link java.beans.BeanInfo BeanInfo}
 * implementations that want adopt the {@link java.beans.Introspector
 * Introspector}'s {@link java.beans.BeanInfo BeanInfo}, selectively
 * replacing properties on demand.
 *
 * Example:
 *
 * <pre>
 *     public class MyBeanBeanInfo
 *         extends DefaultBeanInfo {
 *         
 *         public MyBeanBeanInfo () {
 *             super ();
 *             getPropertyDescriptor ("myProperty").setPropertyEditorClass (MyPropertyEditor.class);
 *         }
 *     }
 * </pre>
 *
 **/
public class DefaultBeanInfo
    implements BeanInfo {

    //----------------------------------------------------------------------
    // constructors
    //----------------------------------------------------------------------
    protected DefaultBeanInfo (Class beanClass, int flags) { setBeanClass (beanClass, flags); }	    
    protected DefaultBeanInfo (Class beanClass)            { setBeanClass (beanClass); }
    protected DefaultBeanInfo () {
	// we are called from sub-class constructor, and the sub-class
	// should be named ...BeanInfo so we can automatically
	// determine the bean class
  	String beanInfoClassName = getClass ().getName ();
	String suffix            = "BeanInfo";
	if (!beanInfoClassName.endsWith (suffix)) {
	    throw new IllegalArgumentException ("illegal class name (" + beanInfoClassName + "); must be of the form <bean class>" + suffix);
	}
	setBeanClass (ClassUtils.forNameCatching (beanInfoClassName.substring (0, beanInfoClassName.length () - suffix.length ())));
    }
	
    //----------------------------------------------------------------------
    private void setBeanInfo (BeanInfo bi) {
// 	System.err.println ("internal bean info: " + bi);
	this.bi = bi;

	ArraysUtils.addAll (eventSetDescriptorMap,
			    bi.getEventSetDescriptors (),
			    ClassUtils.getMethodCatching (EventSetDescriptor.class, "getName"));
	ArraysUtils.addAll (methodDescriptorMap,
			    bi.getMethodDescriptors (),
			    ClassUtils.getMethodCatching (MethodDescriptor.class,   "getName"));
	ArraysUtils.addAll (propertyDescriptorMap,
			    bi.getPropertyDescriptors (),
			    ClassUtils.getMethodCatching (PropertyDescriptor.class, "getName"));

	propertyDescriptors = bi.getPropertyDescriptors ();
    }
    private void setBeanClass (Class beanClass, int flags) {
// 	System.err.println ("setBeanClass (" + beanClass.getName ()+ ", " + flags + ")");
	try {
	    setBeanInfo (Introspector.getBeanInfo (beanClass, flags));
	} catch (IntrospectionException ignore) {
	    // using a SimpleBeanInfo returning null in every method;
	    // thus introspection will be used for every feature and
	    // we will get an IntrospectionException again (but in the
	    // right place)
	    setBeanInfo (new SimpleBeanInfo ());
	}
    }

    private void setBeanClass (Class beanClass) { setBeanClass (beanClass, Introspector.IGNORE_IMMEDIATE_BEANINFO); }

    //----------------------------------------------------------------------
    // interface BeanInfo
    //----------------------------------------------------------------------
    // FIXME:1.3 proxy class?
    public BeanInfo[]           getAdditionalBeanInfo   () 	       { return bi.getAdditionalBeanInfo (); }
    public BeanDescriptor       getBeanDescriptor       () 	       { return bi.getBeanDescriptor (); }
    public int                  getDefaultEventIndex    () 	       { return bi.getDefaultEventIndex (); }
    public int                  getDefaultPropertyIndex () 	       { return bi.getDefaultPropertyIndex (); }
    public EventSetDescriptor[] getEventSetDescriptors  () 	       { return bi.getEventSetDescriptors (); }
    public Image                getIcon                 (int iconKind) { return bi.getIcon (iconKind); }
    public MethodDescriptor[]   getMethodDescriptors    () 	       { return bi.getMethodDescriptors (); }

    private PropertyDescriptor[] propertyDescriptors;
    public PropertyDescriptor[] getPropertyDescriptors  () 	       { return propertyDescriptors; }

    //----------------------------------------------------------------------
    // sub-classes can easily replace items like this:
    //     getPropertyDescriptor ("myProperty").setPropertyEditorClass (MyPropertyEditor.class);
    private Map eventSetDescriptorMap = new LinkedHashMap ();
    public Map getEventSetDescriptorMap () { return eventSetDescriptorMap; }
    public EventSetDescriptor getEventSetDescriptor (String name) { return (EventSetDescriptor)eventSetDescriptorMap.get (name); }

    private Map methodDescriptorMap = new LinkedHashMap ();
    public Map getMethodDescriptorMap () { return methodDescriptorMap; }
    public MethodDescriptor getMethodDescriptor (String name) { return (MethodDescriptor)methodDescriptorMap.get (name); }

    private Map propertyDescriptorMap = new LinkedHashMap ();
    public Map getPropertyDescriptorMap () { return propertyDescriptorMap; }
    public PropertyDescriptor getPropertyDescriptor (String name) { return (PropertyDescriptor)propertyDescriptorMap.get (name); }

    //----------------------------------------------------------------------
    protected void removePropertyDescriptors (Collection removeProperties) {
	Set  removePropertySet      = new HashSet (removeProperties);
	List propertyDescriptorList = new ArrayList (propertyDescriptors.length);
	for (int i = 0; i < propertyDescriptors.length; ++i) {
	    if (removePropertySet.contains (propertyDescriptors [i].getName ())) {
		propertyDescriptorMap.remove (propertyDescriptors [i].getName ());
	    } else {
		propertyDescriptorList.add (propertyDescriptors [i]);
	    }
	}
	propertyDescriptors = (PropertyDescriptor[])propertyDescriptorList.toArray (new PropertyDescriptor [0]);
    }

    //----------------------------------------------------------------------
    private BeanInfo bi;
}
