package com.livis.flabes.util.beans;

import com.livis.flabes.util.*;
import java.awt.Image;
import java.beans.*;
import java.lang.reflect.*;
import java.util.*;

// FIXME: unmodifiable maps (entries delivered by map are modifiable
// as the references point to the original info and thus the array
// view is "updated" automatically), but adding to/removing from a
// map doesn't remove from array, which is counter-intuitive (view
// differ)

/**
 * {@link java.beans.BeanInfo BeanInfo} wrapper class.
 *
 * Objects of this class are instantiated by wrapping existing {@link
 * java.beans.BeanInfo BeanInfo}s and allow for fast (hashed)
 * programmatic access to {@link #getPropertyDescriptor property
 * descriptors}, {@link #getMethodDescriptor method descriptors}, and
 * {@link #getEventSetDescriptor event set descriptors} by name.
 *
 * Example:
 *
 * <pre>
 *     FastBeanInfo       bi = new FastBeanInfo (new MyBeanBeanInfo (), true);
 *     PropertyDescriptor pd = bi.getPropertyDescriptor ("myProperty");
 * </pre>
 *
 **/
public class FastBeanInfo
    implements BeanInfo {

    //----------------------------------------------------------------------
    // constructors
    //----------------------------------------------------------------------
    // unconditionally accept BeanInfo which may have been set
    // manually
    public FastBeanInfo (BeanInfo bi,        boolean rwPropertiesOnly) { setData (bi, rwPropertiesOnly); }
    public FastBeanInfo (BeanInfo bi)                                  { this (bi, false); }
    // lookup map for previously created (class-dependant and thus
    // non-manipulated) BeanInfo
    public FastBeanInfo (Class    beanClass, boolean rwPropertiesOnly) { setData (beanClass, rwPropertiesOnly); }
    public FastBeanInfo (Class    beanClass)                           { this (beanClass, false); }
    // convenience method to free the from calling getClass()
    public FastBeanInfo (Object   bean,      boolean rwPropertiesOnly) { setData (bean.getClass (), rwPropertiesOnly); }
    public FastBeanInfo (Object   bean)                                { this (bean, false); }

    //----------------------------------------------------------------------
    private void setData (BeanInfo bi, boolean rwPropertiesOnly) {
	if (debug) System.err.println ("wrapping BeanInfo for class " + bi.getBeanDescriptor ().getBeanClass ().getName ()
				       + ", " + (rwPropertiesOnly ? "read/write properties only" : "all properties"));
	data = new Data (bi, rwPropertiesOnly);
    }
    private void setData (Class beanClass, boolean rwPropertiesOnly) {
	data = (Data)cache.get (beanClass);
	if (null != data) {
	    if (debug) System.err.println ("reusing cached BeanInfo for class " + beanClass.getName ());
	} else {
	    try {
		data = new Data (Introspector.getBeanInfo (beanClass), rwPropertiesOnly);
		if (debug) System.err.println ("created BeanInfo for class " + beanClass.getName ());
	    } catch (IntrospectionException ignore) {
		// use a SimpleBeanInfo returning null in every
		// method; thus introspection will be used for every
		// feature and we will get an IntrospectionException
		// again (but in the right place)
		data = new Data (new SimpleBeanInfo (), rwPropertiesOnly);
		if (debug) System.err.println ("created SimpleBeanInfo for class " + beanClass.getName ());
	    }
	    cache.put (beanClass, data);
	}
    }

    //----------------------------------------------------------------------
    // interface BeanInfo
    //----------------------------------------------------------------------
    // FIXME:1.3 proxy class?
    public BeanInfo[]           getAdditionalBeanInfo     	() 	       { return data.bi.getAdditionalBeanInfo (); }
    public BeanDescriptor       getBeanDescriptor         	() 	       { return data.bi.getBeanDescriptor (); }
    public int                  getDefaultEventIndex      	() 	       { return data.bi.getDefaultEventIndex (); }
    public int                  getDefaultPropertyIndex   	() 	       { return data.bi.getDefaultPropertyIndex (); }
    public EventSetDescriptor[] getEventSetDescriptors    	() 	       { return data.bi.getEventSetDescriptors (); }
    public Image                getIcon                   	(int iconKind) { return data.bi.getIcon (iconKind); }
    public MethodDescriptor[]   getMethodDescriptors      	() 	       { return data.bi.getMethodDescriptors (); }
    public PropertyDescriptor[] getPropertyDescriptors    	() 	       { return data.getPropertyDescriptors (); }
  	
    public String[]             getPropertyNames          	()             { return data.propertyNames; }
    public Class[]              getPropertyEditorClasses  	()             { return data.getPropertyEditorClasses (); }
    public Class[]              getPropertyElementEditorClasses ()             { return data.getPropertyElementEditorClasses (); }
  
    public PropertyEditor[]     newPropertyEditors              ()             {
	PropertyEditor[] result = new PropertyEditor [getPropertyEditorClasses ().length];
	for (int i = 0; i < result.length; ++i) {
	    result [i] = (PropertyEditor)ClassUtils.newInstanceCatching (getPropertyEditorClasses ()[i], null);
	}
	return result;
    }

    public PropertyEditor[]     newPropertyElementEditors ()             {
	PropertyEditor[] result = new PropertyEditor [getPropertyElementEditorClasses ().length];
	for (int i = 0; i < result.length; ++i) {
	    result [i] = (PropertyEditor)ClassUtils.newInstanceCatching (getPropertyElementEditorClasses ()[i], null);
	}
	return result;
    }

    //----------------------------------------------------------------------
    public Map getEventSetDescriptorMap () { return data.eventSetDescriptorMap; }
    public EventSetDescriptor getEventSetDescriptor (String name) { return (EventSetDescriptor)data.eventSetDescriptorMap.get (name); }

    public Map getMethodDescriptorMap () { return data.methodDescriptorMap; }
    public MethodDescriptor getMethodDescriptor (String name) { return (MethodDescriptor)data.methodDescriptorMap.get (name); }

    public Map getPropertyDescriptorMap () { return data.propertyDescriptorMap; }
    public PropertyDescriptor getPropertyDescriptor (String name) { return (PropertyDescriptor)data.propertyDescriptorMap.get (name); }

    public Set getPropertyNameSet () { return data.propertyNameSet; }

    public Map getPropertyEditorClassMap () { return data.getPropertyEditorClassMap (); }
    public Class getPropertyEditorClass (String name) { return (Class)data.getPropertyEditorClassMap ().get (name); }

    public Map getPropertyElementEditorClassMap () { return data.getPropertyElementEditorClassMap (); }
    public Class getPropertyElementEditorClass (String name) { return (Class)data.getPropertyElementEditorClassMap ().get (name); }

    public Map newPropertyEditorMap () {
	Map              result  = new LinkedHashMap ();
	PropertyEditor[] editors = newPropertyEditors ();
	String[]         names   = getPropertyNames ();
	for (int i = 0; i < editors.length; ++i) {
	    result.put (names [i], editors [i]);
	}
	return result;
    }

    public PropertyEditor newPropertyEditor (String name) { return (PropertyEditor)ClassUtils.newInstanceCatching (getPropertyEditorClass (name), null); }

    public Map newPropertyElementEditorMap () {
	Map              result  = new LinkedHashMap ();
	PropertyEditor[] editors = newPropertyElementEditors ();
	String[]         names   = getPropertyNames ();
	for (int i = 0; i < editors.length; ++i) {
	    result.put (names [i], editors [i]);
	}
	return result;
    }

    public PropertyEditor newPropertyElementEditor (String name) { return (PropertyEditor)ClassUtils.newInstanceCatching (getPropertyElementEditorClass (name), null); }

    //----------------------------------------------------------------------
    private static class Data {
	Data (BeanInfo bi, boolean rwPropertiesOnly) {
	    this.bi = bi;

	    if (rwPropertiesOnly) {
		PropertyDescriptor[] origPDs    = bi.getPropertyDescriptors ();
		List                 newPDList  = new ArrayList (origPDs.length);
		for (int i = 0; i < origPDs.length; ++i) {
		    Method getter = origPDs [i].getReadMethod ();
		    Method setter = origPDs [i].getWriteMethod ();
		    if ((null != getter) && (null != setter)) {
			if (debug) System.err.println ("accepting read/write property " + origPDs [i].getName ());
			newPDList.add (origPDs [i]);
		    } else {
			if (debug) System.err.println ("rejecting non-read/write property " + origPDs [i].getName ());
		    }
		}
		propertyDescriptors = (PropertyDescriptor[])newPDList.toArray (new PropertyDescriptor [0]);
	    } else {
		propertyDescriptors = bi.getPropertyDescriptors ();
	    }

	    ArraysUtils.addAll (eventSetDescriptorMap,
				bi.getEventSetDescriptors (),
				ClassUtils.getMethodCatching (EventSetDescriptor.class, "getName"));
	    ArraysUtils.addAll (methodDescriptorMap,
				bi.getMethodDescriptors (),
				ClassUtils.getMethodCatching (MethodDescriptor.class,   "getName"));
	    ArraysUtils.addAll (propertyDescriptorMap,
				getPropertyDescriptors (),
				ClassUtils.getMethodCatching (PropertyDescriptor.class, "getName"));

	    PropertyDescriptor[] pds = getPropertyDescriptors ();
	    propertyNames            = new String [pds.length];
	    for (int i = 0; i < pds.length; ++i) {
		propertyNames [i] = pds [i].getName ();
		propertyNameSet.add (propertyNames [i]);
	    }
	}

	BeanInfo bi;
	Map 	 eventSetDescriptorMap  = new LinkedHashMap ();
	Map 	 methodDescriptorMap    = new LinkedHashMap ();
	Map 	 propertyDescriptorMap  = new LinkedHashMap ();
	String[] propertyNames;
	Set      propertyNameSet        = new HashSet ();

	private PropertyDescriptor[] propertyDescriptors;
	public PropertyDescriptor[] getPropertyDescriptors () { return propertyDescriptors; }

	//----------------------------------------------------------------------
	// editor creation is expensive, suspend until needed
	//----------------------------------------------------------------------
	private Class[] propertyEditorClasses;
	Class[] getPropertyEditorClasses () {
	    if (null == propertyEditorClasses) {
		propertyEditorClasses = new Class [getPropertyDescriptors ().length];
		for (int i = 0; i < propertyEditorClasses.length; ++i) {
		    PropertyEditor pe = BeanUtils.newEditor (getPropertyDescriptors () [i]);
		    propertyEditorClasses [i] = (null == pe) ? null : pe.getClass ();
		}
	    }
	    return propertyEditorClasses;
	}

	private Map propertyEditorClassMap;
	Map getPropertyEditorClassMap () {
	    if (null == propertyEditorClassMap) {
		propertyEditorClassMap = new LinkedHashMap ();
		// no access to private field, get method creates on
		// demand
		Class[] propertyEditorClasses = getPropertyEditorClasses ();
		for (int i = 0; i < propertyEditorClasses.length; ++i) {
		    propertyEditorClassMap.put (propertyNames [i], propertyEditorClasses [i]);
		}
	    }
	    return propertyEditorClassMap;
	}

	private Class[] propertyElementEditorClasses;
	Class[] getPropertyElementEditorClasses () {
	    if (null == propertyElementEditorClasses) {
		propertyElementEditorClasses = new Class [getPropertyDescriptors ().length];
		for (int i = 0; i < propertyElementEditorClasses.length; ++i) {
		    PropertyEditor pe = BeanUtils.newElementEditor (getPropertyDescriptors () [i]);
		    if (null != pe) {
			if (debug) System.out.println ("INDEXED EDITOR: " + pe);

		    }
		    propertyElementEditorClasses [i] = (null == pe) ? null : pe.getClass ();
		}
	    }
	    return propertyElementEditorClasses;
	}

	private Map propertyElementEditorClassMap;
	Map getPropertyElementEditorClassMap () {
	    if (null == propertyElementEditorClassMap) {
		propertyElementEditorClassMap = new LinkedHashMap ();
		// no access to private field, get method creates on
		// demand
		Class[] propertyElementEditorClasses = getPropertyElementEditorClasses ();
		for (int i = 0; i < propertyElementEditorClasses.length; ++i) {
		    propertyElementEditorClassMap.put (propertyNames [i], propertyElementEditorClasses [i]);
		}
	    }
	    return propertyElementEditorClassMap;
	}
    }

    //----------------------------------------------------------------------
    private        Data    data;
    private static Map     cache = new LinkedHashMap ();
    private static boolean debug = "true".equals (System.getProperty (FastBeanInfo.class.getName () + ".debug"));
}
