package com.livis.flabes.xml;

import com.livis.flabes.util.*;
import java.io.*;
import org.w3c.dom.*;

public class DOMWriter {

    //----------------------------------------------------------------------
    // constructors
    //----------------------------------------------------------------------
    public DOMWriter (String xmlVersion, String encoding, PrintWriter out, Indenter indenter) {
	this.xmlVersion = xmlVersion;
	this.encoding   = encoding;
	this.out        = out;
	this.indenter   = indenter;
    }
    public DOMWriter (                   String encoding, PrintWriter out, Indenter indenter) {
	this (defaultXMLVersion, encoding, out, indenter);
    }
    public DOMWriter (                   String encoding, PrintWriter out) {
	this (defaultXMLVersion, encoding, out, defaultIndenter);
    }
    public DOMWriter (                                    PrintWriter out, Indenter indenter) {
	this (defaultXMLVersion, defaultEncoding, out, indenter);
    }
    public DOMWriter (                                    PrintWriter out) {
	this (defaultXMLVersion, defaultEncoding, out, defaultIndenter);
    }
    public DOMWriter (                                                     Indenter indenter) {
	this (defaultXMLVersion, defaultEncoding, defaultOut, indenter);
    }
    public DOMWriter () {
	this (defaultXMLVersion, defaultEncoding, defaultOut, defaultIndenter);
    }

    private        String      xmlVersion;
    private static String      defaultXMLVersion = "1.0";
    private        String      encoding;
    private static String      defaultEncoding   = null;
    private        PrintWriter out;
    private static PrintWriter defaultOut        = new PrintWriter (System.out);
    private        Indenter    indenter;
    private static Indenter    defaultIndenter   = new Indenter ("", "    ");

    //----------------------------------------------------------------------
    public DOMWriter print (Node node) { return print (node, false); }

    private DOMWriter print (Node node, boolean inRecursion) {
	// !leaf?
	if (node != null) {
	    int type = node.getNodeType ();

	    if (debug) {
		System.err.println ("node: " + node);
		System.err.println ("type: " + type);
	    }

	    switch (type) {
		default: {
		    out.println (indenter + "<!-- unknown node (" + node.getNodeName () + ") of type " + type + "-->");
		} break;

		// print document
		case Node.DOCUMENT_NODE: {
		    out.print (indenter + "<?xml version=\"" + xmlVersion + "\"");
		    if (null != encoding) {
			out.print (" encoding=\"" + encoding + "\"");
		    }
		    out.println ("?>");
		    print (((Document)node).getDocumentElement (), true);
		} break;

		// print element with attributes
		case Node.ELEMENT_NODE: {
		    if (!inRecursion) {
			out.print (indenter);
		    }
		    out.print ("<" + node.getNodeName ());
		    NamedNodeMap attrs = node.getAttributes ();
		    int          len   = attrs.getLength ();
		    for (int i = 0; i < len; i++) {
			Node attr = attrs.item (i);
			out.print (" " +  attr.getNodeName () + "=\"" + attributeConvert (attr.getNodeValue ()) + "\"");
		    }
		    NodeList children = node.getChildNodes ();
		    if ((null == children) || (0 == children.getLength ())) {
			out.print ("/>");
		    } else {
			out.print (">");

			boolean hadNonTextNode = false;

			indenter.incLevel ();

			len = children.getLength ();
			for (int i = 0; i < len; i++) {
			    Node child = children.item (i);
			    if (Node.TEXT_NODE != child.getNodeType ()) {
				hadNonTextNode = true;
				out.println ();
				out.print (indenter);
			    }
			    print (child, true);
			}
			
			indenter.decLevel ();

			if (hadNonTextNode) {
			    out.println ();
			    out.print (indenter);
			}
			out.print ("</" +  node.getNodeName ()+ ">");
		    }
		} break;

		// handle entity reference nodes
		case Node.ENTITY_REFERENCE_NODE: {
		    out.print ("&" + node.getNodeName () + ";");
		} break;

		// print cdata sections
		case Node.CDATA_SECTION_NODE: {
		    out.print (indenter + "<![CDATA[" + node.getNodeValue () + "]]>");
		} break;

		// print text
		case Node.TEXT_NODE: {
		    out.flush ();
		    out.print (tagConvert (node.getNodeValue ()));
		} break;

		// print processing instruction
		case Node.PROCESSING_INSTRUCTION_NODE: {
		    out.print (indenter + "<?" + node.getNodeName ());
		    String data = node.getNodeValue ();
		    if ((null != data) && (0 < data.length ())) {
			out.print (" " +data);
		    }
		    out.print ("?>");
		} break;
	    }
	}
	return this;
    }
    public DOMWriter println () {
	out.println ();
	return this;
    }
    public void flush () {
	out.flush ();
    }

    protected String convert (String s, boolean forAttribute) {
	if (debug) { System.err.println (); System.err.print ("convert (" + s + ", " + forAttribute + "): "); System.err.flush (); }

	if (null == s) {
	    return null;
	}
	
	StringBuffer result = new StringBuffer ();
	int          len    = s.length ();

	if (!forAttribute &&
	    ((256 <= len) || (-1 != s.indexOf ('\n')))) {
	    // long strings and those containing newlines are
	    // formatted as CDATA to ease manual editing
	    result.append ("<![CDATA[" + s + "]]>");
	} else {
	    String       whiteSpaces       = " \t\n\r";
	    boolean      lastWasWhiteSpace = false;
	    for (int i = 0; i < len; i++) {
		char    ch   	        = s.charAt (i);
		
		// for readability, quote space only in a sequence if
		// whitespace characters (next or previous char is
		// white space)
		char    peek       = ((i + 1) < len ) ? s.charAt (i + 1) : /* any non-white space char */ '.';
		boolean quoteSpace = lastWasWhiteSpace || (-1 != whiteSpaces.indexOf (peek));

		lastWasWhiteSpace = false;

		switch (ch) {
		    // standard is not to quote
		    default:   { result.append (ch); } break;

		    // white space (FIXME: Java white space vs. XML
		    // white space vs UNICODE white space; stream
		    // encoding; see
		    // java.lang.Character.isWhitespace()
		    case ' ':  { lastWasWhiteSpace = true; result.append (quoteSpace ? "&#x20;" : " ");  } break;
		    case '\t': { lastWasWhiteSpace = true; result.append ("&#x9;");                      } break;
		    // this cannot happen as strings containing
		    // newlines are formatted as CDATA; leave it here
		    // to be prepared for future changes
		    case '\n': { lastWasWhiteSpace = true; result.append ("&#xA;");                      } break;
		    case '\r': { lastWasWhiteSpace = true; result.append ("&#xD;");                      } break;

		    // general (predefined) entities
		    case '&':  { result.append ("&amp;");  } break;
		    case '<':  { result.append ("&lt;");   } break;
		    case '>':  { result.append ("&gt;");   } break;
		    case '\'': { if (forAttribute) { result.append ("&apos;"); } else { result.append (ch); } } break;
		    case '"':  { if (forAttribute) { result.append ("&quot;"); } else { result.append (ch); } } break;
		}
	    }
	}
	if (debug) System.err.println (result);

	return result.toString ();
    }

    protected String attributeConvert (String s) { return convert (s, true); }
    protected String tagConvert       (String s) { return convert (s, false); }

    //----------------------------------------------------------------------
    private static boolean      debug = "true".equals (System.getProperty (DOMWriter.class.getName () + ".debug"));
}
